/*
 *	adsbox - ADS-B decoder software
 *	(c) atty 2011-17 romulmsg@mail.ru
 */

#include <stdio.h>
#include <getopt.h>
#include <stdlib.h> 
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/timerfd.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <syscall.h>
#include <netinet/tcp.h>
#include "misc.h"
#include "http.h"
#include "adsb.h"
#include "db.h"
#include "file.h"
#include "serial.h"
#include "tcp_avr.h"
#include "tcp_sbs3.h"
#include "tcp_beast.h"
#include "udp_avr.h"

#ifdef RTLSDR
#include "rtlsdr.h"
#endif

#ifdef MLAT
#include "mlat.h"
#endif

#define SBS1_PORT 30003
#define AVR_PORT 30004 
#define HTTP_PORT 8080

pthread_mutex_t mux;
pthread_cond_t cond_var = PTHREAD_COND_INITIALIZER, cond_var_queue =
	PTHREAD_COND_INITIALIZER;
char gdata[BUFLEN];
char grawdata[BUFLEN];
int dmode = 0;
_Bool record = 0;
_Bool seed_server = 0;
#ifdef MLAT
_Bool with_mlat = 0;
#endif
_Bool db_edit = 0;
FILE *rd;
int delay = 0;
sqlite3 *db;

item *head = NULL;
pthread_mutex_t mux_queue;

void enqueue_data(adsb_data *d) {
	item *curr, *tail;

	pthread_mutex_lock(&mux_queue);

	if (head == NULL) {
		head = malloc(sizeof(item));
		memcpy(&head->list_data, d, sizeof(adsb_data));
		head->next = NULL;
	} else {
		curr = head;
		while (curr->next)
			curr = curr->next;
		tail = malloc(sizeof(item));
		curr->next = tail;
		memcpy(&tail->list_data, d, sizeof(adsb_data));
		tail->next = NULL;
	}
	pthread_cond_signal(&cond_var_queue);
	pthread_mutex_unlock(&mux_queue);
}

/* decoder thread */
void *decoder(void *arg) {
	unsigned char msg[128];
	struct tm *now;
	struct timespec ts;
	adsb_data *data;
	item * curr;
	clockid_t cid;
	source_stat *status;

	pthread_getcpuclockid(pthread_self(), &cid);
	db_insert_thread(cid, "ADS-B decoder");

	bzero(msg, sizeof(msg));
	while (1) {

		pthread_mutex_lock(&mux_queue);
		if (head != NULL) {
			data = &head->list_data;
			curr = head;
			head = head->next;
			pthread_mutex_unlock(&mux_queue);
		} else {
			pthread_cond_wait(&cond_var_queue, &mux_queue); // wait signal from sources
			pthread_mutex_unlock(&mux_queue);
			continue;
		}	

		if (data->msg_type == MSG_TYPE_TELEMETRY) { // decoder telemetry packet
			pthread_mutex_lock(&mux);
			sprintf(grawdata, "%s\r\n", data->avr_data);
			pthread_cond_broadcast(&cond_var);
			pthread_mutex_unlock(&mux);
		}

		status = data->status;

		if ((data->msg_type == MSG_TYPE_SHORT && decode_message56(data, (char *) msg))
				|| (data->msg_type == MSG_TYPE_LONG && decode_message(data, (char *) msg))) {
			pthread_mutex_lock(&mux);
			
			status->packets++; // inc packet counter. Bytes counter stats in source handling routines.
			status->bytes += data->data_len;
			
			bzero(grawdata, BUFLEN);
			if (!data->timestamp)
				sprintf(grawdata, "%s\r\n", data->avr_data); // raw data to AVR seed
			else
				sprintf(grawdata, "@%012X%s\r\n", data->timestamp,
						data->avr_data + 1);
			if (record) {
				clock_gettime(CLOCK_REALTIME, &ts);
				now = localtime(&ts.tv_sec);
				fprintf(rd, "%02d:%02d:%02d.%03lu %02d.%02d.%04d %s",
						now->tm_hour, now->tm_min, now->tm_sec,
						ts.tv_nsec / 1000000, now->tm_mday, now->tm_mon + 1,
						now->tm_year + 1900, grawdata);
				fflush(rd);
			}
			bzero(gdata, BUFLEN);
			memcpy(gdata, msg, strlen((char *) msg));
			pthread_cond_broadcast(&cond_var);
			pthread_mutex_unlock(&mux);
			bzero(msg, sizeof(msg));
		}
		free(curr);
	}
	db_delete_thread(cid);
	DEBUG("decoder thread exit.\n")
	pthread_exit(0);
}

/* process client connection thread. MSG format data seed */
void *sbs1_server_client(void *arg) {
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;
	struct timeval timeout;
	timeout.tv_sec = 10;
	timeout.tv_usec = 0;
	int en = 1;

	intptr_t *s = (intptr_t*) arg;

	if (setsockopt(*s, SOL_SOCKET, SO_RCVTIMEO, (char *) &timeout,
			sizeof(timeout)) < 0)
		panic("sbs-1 setsockopt()");
	if (setsockopt(*s, SOL_SOCKET, SO_SNDTIMEO, (char *) &timeout,
			sizeof(timeout)) < 0)
		panic("sbs-1 setsockopt()");
	if (setsockopt(*s, SOL_SOCKET, SO_KEEPALIVE, &en, sizeof(en)) != 0)
		panic("sbs-1 setsockopt()");
	if (setsockopt(*s, IPPROTO_TCP, TCP_NODELAY, &en, sizeof(en)) != 0)
		panic("sbs-1 setsockopt()");

	while (recv(*s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, gdata, strlen(gdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, gdata, strlen(gdata));
			send(*s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(*s);
	free(s);
	DEBUG("\nsbs-1 thread exit");
	pthread_exit(0);
}

/* process client TCP connection thread. AVR data seed */
void *avr_server_client(void *arg) {
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;
	struct timeval timeout;
	timeout.tv_sec = 10;
	timeout.tv_usec = 0;
	int en = 1;

	intptr_t *s = (intptr_t*) arg;

	if (setsockopt(*s, SOL_SOCKET, SO_RCVTIMEO, (char *) &timeout,
			sizeof(timeout)) < 0)
		panic("setsockopt()");
	if (setsockopt(*s, SOL_SOCKET, SO_SNDTIMEO, (char *) &timeout,
			sizeof(timeout)) < 0)
		panic("setsockopt()");
	if (setsockopt(*s, SOL_SOCKET, SO_KEEPALIVE, &en, sizeof(en)) != 0)
		panic("http setsockopt()");
	if (setsockopt(*s, IPPROTO_TCP, TCP_NODELAY, &en, sizeof(en)) != 0)
		panic("http setsockopt()");

	while (recv(*s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, grawdata, strlen(grawdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, grawdata, strlen(grawdata));
			send(*s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(*s);
	free(s);
	DEBUG("AVR seed thread exit\n");
	pthread_exit(0);
}

void usage(void) {
	printf("adsbox [options]\n\t-s, --serial create serial device AVR data source\n");
	printf("\t--baud serial device baudrate (valid rates are: 115200, 230400, 460800,\n");
	printf("\t\t500000, 576000, 921600, 1000000, 1152000, 1500000,\n");
	printf("\t\t2000000, 2500000, 3000000, 3500000, 4000000)\n");
	printf("\t--bits serial device bits (7 or 8)\n");
	printf("\t-d, --daemon be a daemon\n");
	printf("\t-l, --lat data source latitude\n\t-g, --lon data source longitude\n");
	printf("\t--name [name] set data source name\n");
	printf("\t-r, --record [file] record ADS-B data to file\n");
	printf("\t-p, --play [file] play recordered file\n");
	printf("\t-x, --rate [n] play n times faster\n");
	printf("\t-z, --snooze [n] snooze for n seconds before play\n");
	printf("\t--http-port [port] run HTTP server on port (default 8080)\n");
	printf("\t--seed seeding AVR data on TCP port (default 30004)\n");
	printf("\t--seed-port [port] seed AVR data on tcp port\n");
	printf("\t--avr-server [IP] create TCP AVR data source\n");
	printf("\t--avr-server-port [port] connect to AVR data server port (default 30004)\n");
	printf("\t--beast-server [IP] create binary BEAST data source\n");
	printf("\t--beast-server-port [port] connect to BEAST binary data server port (default 10005)\n");
#ifdef RTLSDR
	printf("\t--rtlsdr [=n] create RTLSDR device data source (use device number n, device 0 is default)\n");
	printf("\t--gain [auto|value] set RTLSDR device autogain or gain value (max available by default)\n");
	printf("\t--agc use RTLSDR device AGC\n");
	printf("\t--freq-correct [PPM] set RTLSDR device frequency correction to PPM\n");
	printf("\t--iq-file [file] read SDR IQ samples file\n");
#endif
	printf("\t--sbs3-server [IP] create SBS-3 data source\n");
	printf("\t--sbs3-server-port [port] connect to SBS-3 data server port (default 10001)\n");
	printf("\t--udp-server-port [port] create UDP AVR data source. Listen AVR data on UDP port\n");
	printf("\t--udp-push-to [IP] push AVR data to server\n");
	printf("\t--udp-push-port [port] push AVR data to server port\n");
	printf("\t--dec-threads [n] number of decoder threads (1 by default)\n");
#ifdef MLAT
	printf("\t--mlat perform MLAT (at least 3 data sources are required)\n");
	printf("\t--hz [frequency] data source timestamp counter frequency in Hz units\n");
#endif
	printf("\t--no-events do not log aircraft events (less CPU load)\n");
	printf("\t--no-indirect do not perform indirect calculations of speed, heading etc. (less CPU load)\n");
	printf("\t--db-edit enable write queries in Query Analyzer\n");
	printf("\t-h, --help this help\n");
	exit(0);
}

void sigint_handler() {
	DEBUG("\nSIGINT\n");
	exit(0);
}

void short_timer() {
	static unsigned int count;
	db_stat_thread();
	count += 5;
	if (!(count%12)) {
		db_delete_expired();
		db_delete_expired_log();
		sqlite3_db_release_memory(db);
	}
	return;
}

int main(int ac, char *av[]) {
	extern char * optarg;
	static struct option long_options[] = {
			{"daemon", no_argument, NULL, 'd'},
			{"serial",required_argument, NULL, 's'},
			{"help", no_argument, NULL, 'h'},
			{"lat", required_argument, NULL, 'l'},
			{"lon",required_argument, NULL, 'g'},
			{"record", required_argument, NULL, 'r'},
			{"play", required_argument, NULL, 'p'},
			{"rate", required_argument, NULL, 'x'},
			{"snooze",required_argument, NULL, 'z'},
			{"baud", required_argument, NULL, 1},
			{"bits", required_argument, NULL, 2},
			{"seed", no_argument, NULL, 3},
			{"seed-port", required_argument, NULL, 4},
			{"avr-server", required_argument, NULL, 5},
			{"avr-server-port", required_argument, NULL, 6},
#ifdef RTLSDR
			{"rtlsdr", optional_argument, NULL, 7},
			{"gain", required_argument, NULL, 8},
			{"agc", no_argument, NULL, 9},
			{"freq-correct", required_argument, NULL, 10},
			{"iq-file", required_argument, NULL, 11},
#endif
			{"sbs3-server", required_argument, NULL, 12},
			{"sbs3-server-port", required_argument, NULL, 13},
			{"dec-threads", required_argument, NULL, 14},
			{"udp-push-to", required_argument, NULL, 15},
			{"udp-push-port", required_argument, NULL, 16},
			{"udp-server-port", required_argument, NULL, 17},
			{"beast-server", required_argument, NULL, 18},
			{"beast-server-port", required_argument, NULL, 19},
			{"http-port", required_argument, NULL, 20},
#ifdef MLAT
			{"mlat", no_argument, NULL, 21},
#endif
			{"hz", required_argument, NULL, 22},
			{"no-events", no_argument, NULL, 23},
			{"no-indirect", no_argument, NULL, 24},
			{"name", required_argument, NULL, 25},
			{"db-edit", no_argument, NULL, 26},
			{0, 0, 0, 0} };
	int option_index;
	int pid;
	double playrate = 1;
	long int baudrate = DEFAULT_SERIAL_BAUDRATE;
	long int bits = DEFAULT_SERIAL_BITS;
	int c;
	int rc;
	pthread_t child;
	unsigned int id_source = 0;
	int dec_threads = 1;
	_Bool no_events = 0;
	_Bool no_indirect = 0;

	pthread_mutex_init(&mux, NULL);
	pthread_mutex_init(&mux_queue, NULL);
	udp_push_info *udp_i = NULL;
	
	server_info *hi = malloc(sizeof(server_info));
	hi->port = HTTP_PORT;
	hi->thread_function = http;
	hi->description = "HTTP server\0";

	server_info *ai = malloc(sizeof(server_info));
	ai->port = AVR_PORT;
	ai->thread_function = avr_server_client;
	ai->description = "AVR server\0";

	server_info *si = malloc(sizeof(server_info));
	si->port = SBS1_PORT;
	si->thread_function = sbs1_server_client;
	si->description = "SBS-1 server\0";

	DEBUG("ADSBox ADS-B decoder version %s.\n", VERSION);
	DEBUG("Use --help option for help.\n");
	/*
	 * Dealing with sqlite database.
	 */
	rc = sqlite3_open(":memory:", &db);
	if (rc) {
		DEBUG("Can't open database: %s\n", sqlite3_errmsg(db));
		exit(1);
	}

	if (db_init(db))
		exit(1);
	/*
	 * Done with sqlite.
	 */

	db_insert_thread(CLOCK_PROCESS_CPUTIME_ID, "ADSBox (total)");
	db_insert_thread(CLOCK_THREAD_CPUTIME_ID, "Main");

	while ((c = getopt_long(ac, av, "ds:hl:g:r:p:x:z:", long_options,
			&option_index)) != -1) {
		switch (c) {
		case 1:
			if (!strcmp(optarg, "115200"))
				baudrate = B115200;
			else if (!strcmp(optarg, "230400"))
				baudrate = B230400;
			else if (!strcmp(optarg, "460800"))
				baudrate = B460800;
			else if (!strcmp(optarg, "500000"))
				baudrate = B500000;
			else if (!strcmp(optarg, "576000"))
				baudrate = B576000;
			else if (!strcmp(optarg, "921600"))
				baudrate = B921600;
			else if (!strcmp(optarg, "1000000"))
				baudrate = B1000000;
			else if (!strcmp(optarg, "1152000"))
				baudrate = B1152000;
			else if (!strcmp(optarg, "1500000"))
				baudrate = B1500000;
			else if (!strcmp(optarg, "2000000"))
				baudrate = B2000000;
			else if (!strcmp(optarg, "2500000"))
				baudrate = B2500000;
			else if (!strcmp(optarg, "3000000"))
				baudrate = B3000000;
			else if (!strcmp(optarg, "3500000"))
				baudrate = B3500000;
			else if (!strcmp(optarg, "4000000"))
				baudrate = B4000000;
			else {
				DEBUG("%s - illegal baudrate.\n", optarg);
				exit(1);
			}
			db_update_source_serial_baud(&id_source, &baudrate);
			break;
		case 2:
			if (!strcmp(optarg, "7"))
				bits = CS7;
			else if (!strcmp(optarg, "8"))
				bits = CS8;
			else {
				DEBUG("%s - illegal bits option. Should be 7 or 8.\n", optarg);
				exit(1);
			}
			db_update_source_serial_bits(&id_source, &bits);
			break;
		case 3:
			seed_server = 1; // run AVR server
			break;
		case 4:
			//seed_port = atoi(optarg);
			ai->port = atoi(optarg); // AVR server port
			break;
		case 5:
			id_source++;
			db_insert_source_tcp(&id_source, optarg);
			break;
		case 6:
			db_update_source_tcp_port(&id_source, optarg);
			break;
#ifdef RTLSDR
		case 7:
			id_source++;
			if (optarg != NULL)
				db_insert_source_rtlsdr(&id_source, atoi(optarg));
			else
				db_insert_source_rtlsdr(&id_source, 0);
			break;
		case 8:
			if (!strcmp(optarg, "auto"))
				db_update_source_rtlsdr_gain(&id_source, -2);
			else
				db_update_source_rtlsdr_gain(&id_source, atof(optarg));
			break;
		case 9:
			db_update_source_rtlsdr_agc(&id_source, 1);
			break;
		case 10:
			db_update_source_rtlsdr_freq_correct(&id_source, atoi(optarg));
			break;
		case 11:
			db_update_source_rtlsdr_iq_file(&id_source, optarg);
			break;		
#endif
		case 12:
			id_source++;
			db_insert_source_tcp_sbs3(&id_source, optarg);
			break;
		case 13:
			db_update_source_tcp_sbs3_port(&id_source, optarg);
			break;
		case 14:
			dec_threads = atoi(optarg);
			break;
		case 15:
			udp_i = malloc(sizeof(udp_push_info));
			strncpy(udp_i->addr, optarg, sizeof(udp_i->addr));
			strcpy(udp_i->port, "30005");
			break;
		case 16:
			strncpy(udp_i->port, optarg, sizeof(udp_i->port));
			break;
		case 17:
			id_source++;
			db_insert_source_udp_avr(&id_source, atoi(optarg));
			break;
		case 18:
			id_source++;
			db_insert_source_tcp_beast(&id_source, optarg);
			break;
		case 19:
			db_update_source_tcp_beast_port(&id_source, optarg);
			break;
		case 20:
			hi->port = atoi(optarg);
			break;
#ifdef MLAT
		case 21:
			with_mlat = 1;
			break;
		case 22:
			db_update_hz(id_source, atoi(optarg));
			break;
#endif
		case 23:
			no_events = 1;
			break;
		case 24:
			no_indirect = 1;
			break;
		case 25:
			db_source_name(optarg, &id_source);
			break;	
		case 26:
			db_edit = 1; // enable db write queries in execute_sql()
			break;	
		case 'd':
			dmode = 1;
			break;
		case 's':
			id_source++;
			db_insert_source_serial(&id_source, optarg);
			break;
		case 'l':
			if (db_source_lat(atof(optarg), &id_source) == -1) {
				DEBUG("Can't set latitude for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'g':
			if (db_source_lon(atof(optarg), &id_source) == -1) {
				DEBUG("Can't set longitude for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'r':
			record = 1;
			if ((rd = fopen(optarg, "w+")) == NULL) {
				DEBUG("Unable to open %s\n", optarg);
				exit(1);
			}
			DEBUG("Recording data to %s\n", optarg);
			break;
		case 'p':
			id_source++;
			db_insert_source_file(&id_source, optarg);
			break;
		case 'x':
			playrate = atof(optarg);
			if (db_update_source_file_playrate(&id_source, &playrate)) {
				DEBUG("Can't set playrate for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'z':
			delay = atoi(optarg);
			if (db_update_source_file_delay(&id_source, &delay)) {
				DEBUG("Can't set delay for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'h':
			usage();
			break;
		case '?':
			usage();
			break;
		default:
			DEBUG("Invalid option %s. Try --help for more info.\n", optarg);
			exit(1);
		}
	}

	// log events
	if (!no_events)
		db_evt_trg();
	// perform some calculations
	if (!no_indirect)
		db_indirect_trg();

	if (db_attach("data.sqb", "EXTERNAL"))
			exit(1);

	if (db_init_userdata()) exit(1);

	signal(SIGPIPE, SIG_IGN); // ignoring SIGPIPE causing unexpected process kill
	signal(SIGINT, sigint_handler);

	if (dmode) { /* daemonize */
		printf("Entering daemon mode\n");
		pid = fork();
		if (pid < 0)
			panic("fork()");

		if (pid > 0)
			exit(0);

		umask(0);

		if (setsid() < 0)
			panic("setsid()");

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}
	
	/*
	 * Start ADS-B decoder thread here
	 */
	while (dec_threads--) {
		pthread_create(&child, NULL, &decoder, NULL); /* start decoder thread */
		pthread_detach(child);
	}

	/* start UDP push thread */
	if (udp_i) {
		pthread_create(&child, NULL, &udp_avr_push, udp_i);
		pthread_detach(child);
	}

	if (init_radar_range()) {
		DEBUG("Can't init radar range\n");
		exit(1);
	}

	/* Run data sources */
	run_source_file();
	run_source_tcp();
	run_source_serial();
#ifdef RTLSDR
	run_source_rtlsdr();
#endif
	run_source_tcp_sbs3();
	run_source_tcp_beast();
	run_source_udp_avr();
#ifdef MLAT
	if (with_mlat)
		run_mlat();
#endif
	/* Run servers */
	pthread_create(&child, NULL, &http_server, hi); // http server TODO how to set socket options?
	pthread_detach(child);

	if (seed_server == 1) {
		pthread_create(&child, NULL, &tcp_server, ai); // AVR server
		pthread_detach(child);
	}

	pthread_create(&child, NULL, &tcp_server, si); // SBS-1 server
	pthread_detach(child);

	/*char msg[256];
	adsb_data data;
	data.data_len = 14;
	data.data[0] = 0x8D;
	data.data[1] = 0x15;
	data.data[2] = 0x77;
	data.data[3] = 0x0B;
	data.data[4] = 0x99;
	data.data[5] = 0x04;
	data.data[6] = 0x09;
	data.data[7] = 0x25;
	data.data[8] = 0x48;
	data.data[9] = 0xA0;
	data.data[10] = 0x01; //0x00
	data.data[11] = 0xA4;
	data.data[12] = 0x98;
	data.data[13] = 0xA6;

	data.data_len = 7;
	data.data[0] = 0x5D;
	data.data[1] = 0x40; // 0x42
	data.data[2] = 0x49;
	data.data[3] = 0xC6;
	data.data[4] = 0xF3;
	data.data[5] = 0xFF;
	data.data[6] = 0xDD;

	decode_message(&data, msg);
	exit(0); */

	struct itimerspec new_value;
	struct timespec now;
	int ret;
	unsigned long long val;
	
	if (clock_gettime(CLOCK_REALTIME, &now) == -1) {
		panic("clock_gettime()");
	}

	// create first timer with 5sec interval
	new_value.it_value.tv_sec = now.tv_sec + 1;
	new_value.it_value.tv_nsec = 0;
	new_value.it_interval.tv_sec = 5;
	new_value.it_interval.tv_nsec = 0;

	int fd = timerfd_create(CLOCK_REALTIME, TFD_CLOEXEC);
	if (fd == -1) {
		panic("timerfd_create()");
	}

	if (timerfd_settime(fd, TFD_TIMER_ABSTIME, &new_value, NULL) == -1) {
		panic("timerfd_settime()");
	}

	int max_d = fd;
	max_d++;

	fd_set input;
	
	// process timers here
	while (1) {
		FD_ZERO(&input);
		FD_SET(fd, &input);

		if (select(max_d, &input, NULL, NULL, NULL) > 0) {
			
			if (FD_ISSET(fd, &input)) {
				ret = read(fd, &val, sizeof(val));
				if (ret == -1)
				    panic("read()");
				short_timer();
			}
		}
	}
	exit(0);
}
